# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Launcher for fumi.

This parses the command line arguments and performs the different operations.
"""

__version__ = '0.4.0'

# Set locale
from pkg_resources import resource_filename
import six
import gettext

_locale_dir = resource_filename(__name__, 'locale')
if six.PY3:
    gettext.install('fumi', _locale_dir)
else:
    gettext.install('fumi', _locale_dir, unicode=True)


import argparse
import os
import sys

from fumi import messages as m
from fumi import util
from fumi.deployer import build_deployer


FUMI_YML = os.path.join(os.getcwd(), 'fumi.yml')


def deploy(conf_name, prepare=False):
    """Deploy using given configuration.

    Arguments:
        conf_name (str): Name of the configuration to use in the deployment.
        prepare (bool): Whether or not this is a preparation deployment.
            A preparation simply checks connection and creates the remote
            directory tree.
    """
    status, content = util.read_yaml(FUMI_YML)

    if not status:
        sys.exit(-1)

    if not content:
        util.cprint(m.NO_YML, 'red')
        sys.exit(-1)

    if not conf_name:
        # Find default configuration
        default = None

        for k in content.keys():
            if content[k].get('default'):
                # Found default configuration
                default = k
                util.cprint(m.USE_DEFAULT_CONF % default)
                break

        if not default:
            # Default not found
            if len(content.keys()) == 0:
                util.cprint(m.NO_CONFS, 'red')
                sys.exit(-1)

            # Ask for default
            util.cprint(m.CONFS_FOUND)

            for k in content.keys():
                util.cprint('- %s' % k)

            default = six.moves.input(m.CONF_SET_DEF)

            if default in content.keys():
                content[default]['default'] = True
                util.write_yaml(FUMI_YML, content)

            else:
                # Welp...
                util.cprint(m.CONF_NOT_FOUND, 'red')
                sys.exit(-1)

        conf_name = default

    elif conf_name not in content.keys():
        util.cprint(m.CONF_NAME_NOT_FOUND % conf_name, 'red')
        sys.exit(-1)

    # Build deployer
    status, deployer = build_deployer(content[conf_name])

    if not status:
        sys.exit(-1)

    if prepare:
        # Preparation
        result = deployer.prepare()

    else:
        # Deploy!
        result = deployer.deploy()

    if not result:
        sys.exit(-1)


def list_configs():
    """List the configurations present in the fumi.yml file."""
    status, content = util.read_yaml(FUMI_YML)

    if not status:
        sys.exit(-1)

    if not content:
        util.cprint(m.NO_YML, 'red')
        sys.exit(-1)

    for conf in content.keys():
        is_default = content[conf].get('default', False)

        if is_default:
            util.cprint(m.LIST_DEFAULT % conf)

        else:
            util.cprint('- %s' % conf)


def new_config(name):
    """Create new basic configuration in fumi.yml file.

    Arguments:
        name (str): Name for the new configuration.
    """
    status, content = util.read_yaml(FUMI_YML)

    if not status:
        sys.exit(-1)

    if not content:
        # Starting from scratch
        content = {}

    if name in content.keys():
        # Do not overwrite configuration
        util.cprint(m.CONF_EXISTS % name, 'red')
        sys.exit(-1)

    content[name] = {
        'source-type' : '',
        'source-path' : '',

        'predep'  : [],
        'postdep' : [],

        'host'         : '',
        'user'         : '',
        'use-password' : False,
        'password'     : '',
        'deploy-path'  : '',
    }

    status = util.write_yaml(FUMI_YML, content)
    if not status:
        sys.exit(-1)

    util.cprint(m.CREATED_BLANK % name)


def remove_config(name):
    """Remove a configuration from the fumi.yml file.

    Arguments:
        name (str): Name of the configuration to remove.
    """
    status, content = util.read_yaml(FUMI_YML)

    if not status:
        sys.exit(-1)

    if not content:
        util.cprint(m.NO_YML, 'red')
        sys.exit(-1)

    if name not in content.keys():
        util.cprint(m.CONF_NAME_NOT_FOUND % name, 'red')
        sys.exit(-1)

    # Remove section
    del content[name]

    status = util.write_yaml(FUMI_YML, content)
    if not status:
        sys.exit(-1)

    util.cprint(m.CONF_REMOVED % name, 'green')


def init_parser():
    """ Initialize the arguments parser. """
    parser = argparse.ArgumentParser(description=m.FUMI_DESC)

    parser.add_argument('--version', action='version',
        version='%(prog)s ' + __version__)

    subparsers = parser.add_subparsers(title=m.FUMI_CMDS)


    # deploy
    parser_deploy = subparsers.add_parser('deploy', help=m.FUMI_DEPLOY_DESC)
    parser_deploy.add_argument(
        'configuration',
        nargs='?',
        metavar=m.FUMI_CONF,
        help=m.FUMI_CONF_DESC
    )


    # list
    parser_list = subparsers.add_parser('list', help=m.FUMI_LIST_DESC)


    # new
    parser_new = subparsers.add_parser('new', help=m.FUMI_NEW_DESC)
    parser_new.add_argument(
        'name',
        metavar=m.FUMI_NAME,
        help=m.FUMI_NAME_DESC
    )


    # prepare
    parser_prepare = subparsers.add_parser('prepare', help=m.FUMI_PREP_DESC)
    parser_prepare.add_argument(
        'configuration',
        nargs='?',
        metavar=m.FUMI_CONF,
        help=m.FUMI_CONF_DESC
    )


    # remove
    parser_remove = subparsers.add_parser('remove', help=m.FUMI_RM_DESC)
    parser_remove.add_argument(
        'name',
        metavar=m.FUMI_NAME,
        help=m.FUMI_NAME_DESC
    )

    return parser

def parse_action(action, parsed):
    """ Parse the action to execute. """
    if action == 'deploy':
        deploy(parsed.configuration)

    elif action == 'list':
        list_configs()

    elif action == 'new':
        new_config(parsed.name)

    elif action == 'prepare':
        deploy(parsed.configuration, prepare=True)

    elif action == 'remove':
        remove_config(parsed.name)

    else:
        util.cprint(m.FUMI_UNKNOWN)

def main():
    """Entrypoint for the program."""
    parser = init_parser()
    args = parser.parse_args()

    if len(sys.argv) == 1:
        # No argument provided
        parser.print_help()
        return

    parse_action(sys.argv[1], args)



# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Message strings."""

import gettext

CMD_EXEC = _('Command execution')
CMD_LOCAL = _('Running local command: %s')
CMD_REMOTE = _('Running remote command: %s')
CMD_SKIP = _('Skipping unknown command type "%s"')

# NOTE: The named configuration already exists in the file
CONF_EXISTS = _('Configuration "%s" already exists')
# NOTE: The chosen configuration does not exist
CONF_NAME_NOT_FOUND = _('Configuration "%s" does not exist')
CONF_NOT_FOUND = _('That configuration does not exist!')
# NOTE: Name of configuration that was removed
CONF_REMOVED = _('Removed configuration: %s')
# NOTE: Shows list of configuration names to user
CONF_SET_DEF = _('Which one do you want to set as default?: ')
# NOTE: List of configurations after scanning the YML file
CONFS_FOUND = _('I found the following configurations:')

CONN_AUTH_FAIL = _('Authentication failed')
CONN_FAIL = _('Could not connect, please check your credentials')
CONN_NEEDPASS = _('Connection needs a password')
CONN_KEYPASS = _('Password needed to unlock private key file')
# NOTE: When introducing password manually
CONN_PASS = _('Password: ')
CONN_PUBKEY = _('Trying to connect using public key')
CONN_TRYPASS = _('Trying to connect using provided password')

CORRECT = _('Correct!')

# NOTE: Created empty configuration template, shows name
CREATED_BLANK = _('Created blank configuration: %s')

DEP_COMPLETE = _('Deployment complete!')
DEP_CHECK_REMOTE = _('Checking remote directories...')
DEP_CONNECTED = _('Connected!')
DEP_CONNECTING = _('Connecting to %s as %s...')
DEP_CREATEDIR = _('Creating remote directory tree...')
DEP_GIT = _('Performing a "git" deployment')
DEP_GIT_CLONE = _('Cloning repository...')
DEP_GIT_NOTFOUND = _('git command not found in remote server')
DEP_LOCAL = _('Performing a "local"deployment')
DEP_LOCAL_CLEAN = _('Cleaning temporary files...')
DEP_LOCAL_COMPRESS = _('Compressing source to %s')
DEP_LOCAL_ERR1 = _('Error: some files differ')
DEP_LOCAL_ERR127 = _('Error: tar command not found in remote host')
DEP_LOCAL_ERR2 = _('Fatal error when extracting remote file')
DEP_LOCAL_PATHNOEXIST = _('Path "%s" does not exist, ignoring')
DEP_LOCAL_SCPFAIL = _('Failed to initiate SCP, check configuration')
DEP_LOCAL_UNCOMPRESS = _('Uncompressing remote file...')
DEP_LOCAL_UPLOAD = _('Uploading %s...')
DEP_LOCAL_UPLOADERR = _('Error uploading to server: %s')
DEP_MISSING_PARAM = _('Missing required parameter: %s')
DEP_PREPARE_COMPLETE = _('Preparation complete!')
DEP_PREPARE_REV = _('Preparing revision %s')
DEP_PREPARE_NOTICE = _('Make sure to upload shared files before deploying')
DEP_UNKNOWN = _('Unknown deployment type: %s')

DONE = _('Done!')

# NOTE: Command line title for the commands section
FUMI_CMDS = _('commands')
FUMI_CONF = _('configuration')
FUMI_CONF_DESC = _('configuration to use')
FUMI_DEPLOY_DESC = _('deploy using given configuration')
FUMI_DESC = _('Simple deployment tool')
FUMI_LIST_DESC = _('list all the available deployment configurations')
FUMI_NAME = _('name')
FUMI_NAME_DESC = _('name for the new configuration')
FUMI_NEW_DESC = _('create new deployment configuration')
FUMI_PREP_DESC = _('test connection and prepare remote directories')
FUMI_RM_DESC = _('remove a configuration from the deployment file')
# TODO: Shown when trying to execute an unknown action
FUMI_UNKNOWN = _('Unknown action')

LINK_DIR = _('Linking directory...')
LINK_SHARED = _('Linking shared files...')
LINKING = _('Linking: %s')

# NOTE: When listing configurations, the default one is marked as such
LIST_DEFAULT = _('- %s (default)')
NO_CONFS = _('There are no configurations available')
NO_YML = _('There is no fumi.yml file')

PATH_NOEXIST = _('Path "%s" does not exist')

REMOTE_DEP_CREATE_ERR = _('Cannot create remote deployment directory')
REMOTE_DEP_NOEXIST = _('Remote deployment directory does not exist')
REMOTE_FILE_NOEXIST = _('Remote file "%s" does not exist')
REMOTE_REV_CREATE_ERR = _('Cannot create remote revisions directory')
REMOTE_REV_NOEXIST = _('Remote revisions directory does not exist')
REMOTE_SHR_CREATE_ERR = _('Cannot create remote shared directory')
REMOTE_SHR_NOEXIST = _('Remote shared directory does not exist')
REMOTE_TMP_CREATE_ERR = _('Cannot create remote temporary directory')
REMOTE_TMP_NOEXIST = _('Remote temporary directory does not exist')

REV_CHECK = _('Checking old revisions...')
REV_LINK_PREV = _('Linking previous revision (%s) ...')
REV_LIST_ERR = _('Error obtaining list of revisions')
REV_PREV_MISSING = _('No previous revision to link')
# NOTE: ID of the revision being removed
REV_RM = _('Removing revision %s')
REV_RM_REMOTE = _('Removing remote revision %s ...')

# NOTE: First token is the path, the second is the exception
RM_ERR = _('Could not remove "%s": %s')
# NOTE: Next lines are directly read from stderr
RM_ERR_REMOTED = _('Remote error, could not remove directory:')
# NOTE: Next lines are directly read from stderr
RM_ERR_REMOTEF = _('Remote error, could not remove file:')
# NOTE: Only shows filename
RM_FAIL = _('Failed to remove file: %s')
# NOTE: Path being removed
RM_MSG_LOCAL = _('Removing %s ...')
RM_MSG_REMOTE = _('Removing remote file %s ...')

ROLLBACK_BEGIN = _('Beginning rollback...')

# TODO: Includes exception message
UNEXPECTED_ERR = _('Unexpected error: %s')

# NOTE: Token represents name of the default configuration
USE_DEFAULT_CONF = _('Using default configuration: %s')

# NOTE: Includes the error message
YML_ERR = _('Error in deployment file: %s')
# NOTE: Includes the error message
YML_WRITE_ERR = _('Error writing yaml to deployment file: %s')



# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Utility code for fumi.

These are common functions used in different deployments.
"""

from __future__ import print_function

import blessings
import getpass
import gettext
import os
import paramiko
import shutil
import subprocess
import yaml

from fumi import messages as m

COLOR_TERM = blessings.Terminal()


def check_dirs(ssh, deployer):
    """Check if all the necessary directories exist in the remote host.

    Arguments:
        ssh: Established SSH connection instance.
        deployer (``Deployer``): Deployer instance.

    Returns:
        Boolean indicating result.
    """
    # Remote temporary
    if not dir_exists(ssh, deployer.host_tmp):
        cprint(m.REMOTE_TMP_NOEXIST, 'red')
        return False

    # Deployment path
    if not dir_exists(ssh, deployer.deploy_path):
        cprint(m.REMOTE_DEP_NOEXIST, 'red')
        return False

    # Revisions
    rev = os.path.join(deployer.deploy_path, 'rev')
    if not dir_exists(ssh, rev):
        cprint(m.REMOTE_REV_NOEXIST, 'red')
        return False

    # Shared files
    shared = os.path.join(deployer.deploy_path, 'shared')
    if not dir_exists(ssh, shared):
        cprint(m.REMOTE_SHR_NOEXIST, 'red')
        return False

    return True

def create_dirs(ssh, deployer):
    """Create remote directories.

    Arguments:
        ssh: Established SSH connection instance.
        deployer (``Deployer``): Deployer instance.

    Returns:
        Boolean indicating result.
    """
    # Remote temporary
    if deployer.host_tmp != '/tmp' and not create_tree(ssh, deployer.host_tmp):
        cprint(m.REMOTE_TMP_CREATE_ERR, 'red')
        return False

    # Deployment path
    if not create_tree(ssh, deployer.deploy_path):
        cprint(m.REMOTE_DEP_CREATE_ERR, 'red')
        return False

    # Revisions
    rev = os.path.join(deployer.deploy_path, 'rev')
    if not create_tree(ssh, rev):
        cprint(m.REMOTE_REV_CREATE_ERR, 'red')
        return False

    # Shared files
    shared = os.path.join(deployer.deploy_path, 'shared')
    if not create_tree(ssh, shared):
        cprint(m.REMOTE_SHR_CREATE_ERR, 'red')
        return False

    return True

def connect(deployer):
    """Try to connect to the remote host through SSH.

    Arguments:
        dep (``Deployer``): Deployer instance.

    Returns:
        Boolean indicating result and paramiko.SSHClient instance or ``None``.
    """
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    if not deployer.use_password:
        # Not using password, rely on public key authentication
        try:
            cprint(m.CONN_PUBKEY + '\n', 'magenta')
            ssh.connect(deployer.host, username=deployer.user)

        except paramiko.ssh_exception.AuthenticationException:
            cprint(m.CONN_AUTH_FAIL, 'red')
            return False, None

        except paramiko.ssh_exception.PasswordRequiredException:
            cprint(m.CONN_KEYPASS, 'red')
            return False, None

        except Exception as e:
            # Unknown exception
            cprint(m.UNEXPECTED_ERR % e, 'red')
            return False, None

    elif deployer.use_password and not deployer.password:
        # Using password but not supplied in the yml file, ask for it
        cprint(m.CONN_NEEDPASS + '\n', 'magenta')
        pwd = getpass.getpass(m.CONN_PASS)

        try:
            ssh.connect(deployer.host, username=deployer.user, password=pwd)

        except:
            cprint(m.CONN_FAIL, 'red')
            return False, None

    elif deployer.use_password and deployer.password:
        # Using password supplied in the yml file
        cprint(m.CONN_TRYPASS + '\n' , 'magenta')

        try:
            ssh.connect(
                deployer.host,
                username=deployer.user,
                password=deployer.password)

        except:
            cprint(m.CONN_FAIL, 'red')
            return False, None

    return True, ssh

def clean_revisions(ssh, keep_max, rev_path):
    """Remove old revisions from the remote server.

    Only the most recent revisions will be kept.

    Arguments:
        ssh: Established SSH connection instance.
        max_keep (int): Maximum number of revisions to keep.
        rev_path (str): Path to the revisions directory in remote host.

    Returns:
        Boolean indicating result.
    """
    cprint('> ' + m.REV_CHECK, 'cyan')

    stdin, stdout, stderr = ssh.exec_command('ls %s' % rev_path)
    status = stdout.channel.recv_exit_status()

    if status == 1 or status == 2:
        cprint(m.REV_LIST_ERR, 'red')
        cprint(*stderr.readlines())
        return False

    revisions = stdout.readlines()

    if len(revisions) > keep_max:
        old_revisions = []

        while len(revisions) > keep_max:
            old_revisions.append(revisions.pop(0))

        for r in old_revisions:
            cprint(m.REV_RM % r.strip(), 'magenta')
            rm_old = 'rm -rf %s' % os.path.join(rev_path, r.strip())
            stdin, stdout, stderr = ssh.exec_command(rm_old)

    cprint(m.DONE +'\n', 'green')
    return True

def create_tree(ssh, path):
    """Try to create a remote tree path.

    Arguments:
        ssh: Established SSH connection instance.
        path (str): Tree path to create.

    Returns:
        Boolean indicating whether the tree was created or not.
    """
    stdin, stdout, stderr = ssh.exec_command(
        "mkdir -p %s" % path)

    if stdout.channel.recv_exit_status() == 0:
        return True

    return False

def dir_exists(ssh, path):
    """Check whether a remote directory exists or not.

    Arguments:
        ssh: Established SSH connection instance.
        path (str): Path of the directory to check.

    Returns:
        Boolean indicating if the directory exists or not.
    """
    stdin, stdout, stderr = ssh.exec_command(
        "[ -d %s ] && echo OK" % path)

    result = stdout.read()

    if result:
        return True

    return False

def cprint(text, color='normal', bold=True):
    """Print the given text using blessings terminal.

    Arguments:
        text (str): Text to print.
        color (str): Color to print the message in.
        bold (bool): Whether or not the text should be printed in bold.
    """
    if color == 'cyan':
        to_print = COLOR_TERM.cyan(text)

    elif color == 'green':
        to_print = COLOR_TERM.green(text)

    elif color == 'magenta':
        to_print = COLOR_TERM.magenta(text)

    elif color == 'red':
        to_print = COLOR_TERM.red(text)

    elif color == 'white':
        to_print = COLOR_TERM.white(text)

    else:
        # Normal text
        to_print = text

    if bold and color != 'normal':
        print(COLOR_TERM.bold(to_print))

    else:
        print(to_print)

def read_yaml(path):
    """Reads the given YAML file.

    Arguments:
        path (str): Path to the file to read (fumi.yml).

    Returns:
        Boolean indicating result and ``dict`` with the information or ``None``.
    """
    if os.path.isfile(path):
        with open(path, 'r') as fumi_yml:
            try:
                content = yaml.load(fumi_yml)

            except yaml.YAMLError as e:
                cprint(m.YML_ERR % e, 'red')
                return False, None

        return True, content

    return True, None

def remove_local(path):
    """Remove a local file or directory.

    Arguments:
        path (str): Absolute path to the file or directory.

    Returns:
        Boolean indicating result.
    """
    if os.path.isfile(path):
        # Regular file
        remover = os.remove

    elif os.path.isdir(path):
        # Directory
        remover = shutil.rmtree

    else:
        # What?
        cprint(m.PATH_NOEXIST % path, 'red')
        return False

    try:
        remover(path)

    except Exception as e:
        # Something failed
        cprint(m.RM_ERR % (path, e), 'red')
        return False

    return True

def remove_remote(ssh, path):
    """Remove a remote file or directory.

    Arguments:
        ssh: Established SSH connection instance.
        path (str): Absolute path to the file or directory.

    Returns:
        Boolean indicating result.
    """
    stdin, stdout, stderr = ssh.exec_command('rm -rf %s' % path)

    return True

def rollback(ssh, deployer, timestamp, level):
    """Perform a rollback based on current deployment.

    Rollbacks have several levels:

        1. Remove locally generated files (compressed source).
        2. Remove uploaded compressed source (if any).
        3. Remove remote revision.
        4. Link previous revision.

    All the levels lower to the one provided are executed as well.

    Arguments:
        ssh: Established SSH connection instance.
        deployer (``Deployer``): Deployer instance.
        timestamp (str): Timestamp for the revision to rollback.
        level (int): Level of rollback to perform.

    Returns:
        Boolean indicating result of the rollback.
    """
    cprint('> ' + m.ROLLBACK_BEGIN, 'cyan')

    # Relevant names
    comp_file = timestamp + '.tar.gz'
    rev_path = os.path.join(deployer.deploy_path, 'rev')

    if level >= 1:
        # Remove local files
        local_file = os.path.join('/tmp', comp_file)

        if os.path.isfile(local_file):
            cprint(m.RM_MSG_LOCAL % local_file, 'magenta')

            try:
                os.remove(local_file)

            except:
                # Do not exit because of local file error, simply ignore
                cprint(m.RM_FAIL % local_file, 'red')
                pass

    if level >= 2:
        # Remove remote files
        uload_tmp = deployer.host_tmp or '/tmp'
        remote_file = os.path.join(uload_tmp, comp_file)

        # Check if file exists
        stdin, stdout, stderr = ssh.exec_command(
            '[ -f %s ] && echo OK' % remote_file)

        result = stdout.read()

        if result:
            # File exists
            cprint(m.RM_MSG_REMOTE % remote_file, 'magenta')

            stdin, stdout, stderr = ssh.exec_command(
                'rm %s' % remote_file)

            status = stdout.channel.recv_exit_status()

            if status < 0:
                # Could not remove it?
                cprint(m.RM_ERR_REMOTEF, 'red')
                cprint(*stderr.readlines())

        else:
            # File does not exist
            cprint(m.REMOTE_FILE_NOEXIST % remote_file, 'white')

    if level >= 3:
        # Remove revision
        stdin, stdout, stderr = ssh.exec_command(
            '[ -d %s ] && echo OK' % rev_path)

        result = stdout.read()

        if result:
            # Directory exists
            cprint(m.REV_RM_REMOTE % timestamp, 'magenta')

            stdin, stdout, stderr = ssh.exec_command(
                'rm -rf %s' % os.path.join(rev_path, timestamp))

            status = stdout.channel.recv_exit_status()

            if status < 0:
                cprint(m.RM_ERR_REMOTED, 'red')
                cprint(*stderr.readlines())

        else:
            # File does not exist
            cprint(m.REMOTE_REV_NOEXIST, 'white')

    if level >= 4:
        # Link previous version
        stdin, stdout, stderr = ssh.exec_command('ls %s' % rev_path)
        status = stdout.channel.recv_exit_status()

        revs = [r.rstrip() for r in stdout.readlines()]

        if timestamp in revs:
            revs.remove(timestamp)

        if len(revs) > 0:
            cprint(m.REV_LINK_PREV % revs[-1], 'magenta')

            link_path = os.path.join(deployer.deploy_path, 'current')
            previous_rev = os.path.join(deployer.deploy_path, 'rev', revs[-1])

            ln = 'ln -sfn %s %s' % (previous_rev, link_path)

            stdin, stdout, stderr = ssh.exec_command(ln)

        else:
            # No more revisions
            cprint(m.REV_PREV_MISSING, 'white')

    cprint(m.DONE, 'green')
    return True

def run_commands(ssh, commands, remote_path=None):
    """Execute pre and post deployment commands (both local and remote).

    Remote pre-deployment commands are usually executed in the user's
    directory (~/) when possible, while post-deployment commands are executed
    in the directory of the revision that has been deployed
    (deploy_path/current).

    Arguments:
        ssh: Established SSH connection instance.
        commands (list[str]): List of commands to execute (predep or postdep).
        remote_path (str): Remote path in which to execute commands.

    Returns:
        Boolean indicating result.
    """
    if not commands:
        # No commands to execute
        return True

    cprint('> ' + m.CMD_EXEC, 'cyan')

    popd = 'popd > /dev/null 2>&1'

    for cmd in commands:
        command_type = cmd[0]
        to_run = cmd[1]

        if command_type == 'local':
            cprint(m.CMD_LOCAL % to_run, 'magenta')

            subprocess.Popen(to_run, shell=True).wait()

        elif command_type == 'remote':
            cprint(m.CMD_REMOTE % to_run, 'magenta')

            if remote_path:
                # Only for post-deployment commands
                pushd = 'pushd %s > /dev/null 2>&1' % remote_path
                to_run = '%s; %s; %s' % (pushd, to_run, popd)

            stdin, stdout, stderr = ssh.exec_command(to_run)

            # Print command output in real-time
            while not stdout.channel.exit_status_ready():
                if stdout.channel.recv_ready():
                    cprint(stdout.readline())

        else:
            # Unknown type, skip
            cprint((m.CMD_SKIP % command_type) + '\n', 'red')

    cprint(m.DONE + '\n', 'green')
    return True

def symlink(ssh, deployer, rev_path, timestamp):
    """Symlink the deployed revision to the deploy_path/current directory.

    Arguments:
        ssh: Established SSH connection instance.
        deployer (``Deployer``): Deployer instance.
        rev_path (str): Path to the revisions directory in remote host.
        timestamp (str): Timestamp that identifies current revision.

    Returns:
        Boolean indicating result.
    """
    cprint('> ' + m.LINK_DIR, 'cyan')

    link_path = os.path.join(deployer.deploy_path, 'current')
    current_rev = os.path.join(rev_path, timestamp)

    ln = 'ln -sfn %s %s' % (current_rev, link_path)

    stdin, stdout, stderr = ssh.exec_command(ln)
    # status = stdout.channel.recv_exit_status()

    cprint(m.DONE + '\n', 'green')
    return True

def symlink_shared(ssh, deployer):
    """Symlink shared files to the deploy_path/current directory.

    Arguments:
        ssh: Established SSH connection instance.
        deployer (``Deployer``): Deployer instance.

    Returns:
        Boolean indicating result.
    """
    if not deployer.shared_paths:
        # Nothing to link
        return True

    cprint('> ' + m.LINK_SHARED, 'cyan')

    current_path = os.path.join(deployer.deploy_path, 'current')
    shared_path = os.path.join(deployer.deploy_path, 'shared')

    for shared in deployer.shared_paths:
        cprint(m.LINKING % shared, 'magenta')

        src_path = os.path.join(shared_path, shared)
        dest_path = os.path.join(current_path, shared)

        ln = 'ln -sfn %s %s' % (src_path, dest_path)

        stdin, stdout, stderr = ssh.exec_command(ln)
        # status = stdout.channel.recv_exit_status()

    cprint(m.DONE + '\n', 'green')
    return True

def write_yaml(path, content):
    """Overwrite the content of the given YAML file.

    Arguments:
        path (str): Path to the file (fumi.yml).
        content (str): Content to write.

    Returns:
        Boolean indicating result.
    """
    with open(path, 'w') as fumi_yml:
        try:
            yaml.dump(content, fumi_yml, default_flow_style=False)

        except yaml.YAMLError as e:
            cprint(m.YML_WRITE_ERR % e, 'red')
            return False

    return True



# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from babel.messages import frontend as babel
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "DESCRIPTION.rst")) as f:
    long_description = f.read()

setup(
    name='fumi',
    version='0.4.0',

    description='A small and (hopefully) simple deployment tool',
    long_description=long_description,

    url='https://github.com/rmed/fumi',

    author='Rafael Medina García',
    author_email='rafamedgar@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='fumi deploy git remote ssh',

    packages=find_packages(),

    package_data = {
        'fumi': ['locale/*/LC_MESSAGES/*.mo']
    },

    exclude_package_data = {
        '': ['*.po', '*.pot']
    },

    cmdclass = {
        'compile_catalog': babel.compile_catalog,
        'extract_messages': babel.extract_messages,
        'init_catalog': babel.init_catalog,
        'update_catalog': babel.update_catalog
    },

    install_requires=[
        'blessings==1.6',
        'paramiko==2.0.2',
        'pyyaml==3.12',
        'scp==0.10.2',
        'six==1.10.0'
    ],

    entry_points={
        'console_scripts': [
            'fumi = fumi.launcher:main'
        ]
    }
)



# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Code for the ``Deployer`` class, which acts as proxy for configurations."""

import gettext
import types

from fumi import messages as m
from fumi import deployments
from fumi.util import cprint


class Deployer(object):
    """Configuration parsed from the ``fumi.yml`` file.

    Attributes:
        source_type (str): Source type (e.g. 'local' or 'git'). Required.
        source_path (str): Path to the source files in local machine. Required.
        host (str): Host to perform the deployment in. Required.
        user (str): User to use for the deployment. Required.
        use_password (bool): Whether or not to use password. If set to ``False``
            (default), will rely on public key authentication. Otherwise, it
            will be asked for during deployment or may be configured through the
            ``password`` attribute.
        password (str): If ``use_password`` is set to True, this attribute can
            be used to specify the password used for the connection. Otherwise
            it will be asked for during deployment.
        deploy_path (str): Remote host path in which to deploy files. Required.
        predep (list[str]): List of commands to execute before deploying.
        postdep (list[str]): List of commands to execute after deploying.
        host_tmp (str): In ``local`` deployments, the remote directory to use
            for uploading the compressed files (defaults to ``'/tmp'``).
        keep_max (int): Maximum revisions to keep in the remote server.
        local_ignore (list[str]): List of files (or directories) to ignore in
            ``local`` deployments.
        buffer_size (int): Buffer size (in bytes) for file copying in ``local``
            deployments. Defaults to 1 MB.
        shared_paths (list[str]): List of file and directory paths that
            should be shared accross deployments. These are relative to the
            root of the project and are linked to the current revision.
    """

    def __init__(self, **kwargs):
        # Source information
        self.source_type = kwargs['source-type']
        self.source_path = kwargs['source-path']

        # Destination host information
        self.host = kwargs['host']
        self.user = kwargs['user']
        self.use_password = kwargs.get('use-password', False)
        self.password = kwargs.get('password')
        self.deploy_path = kwargs['deploy-path']

        # Pre-deployment commands
        predep = kwargs.get('predep', [])
        self.predep = []

        for p in predep:
            # Single key dicts
            for k, v in p.items():
                self.predep.append((k, v))

        # Post-deployment commands
        postdep = kwargs.get('postdep', [])
        self.postdep = []

        for p in postdep:
            # Single key dicts
            for k, v in p.items():
                self.postdep.append((k, v))


        # Optional information
        self.host_tmp = kwargs.get('host-tmp', '/tmp')
        self.keep_max = kwargs.get('keep-max')
        self.local_ignore = kwargs.get('local-ignore')
        self.buffer_size = int(kwargs.get('buffer-size', 1024 * 1024))
        self.shared_paths = kwargs.get('shared-paths', [])

def build_deployer(config):
    """Build a Deployer object.

    Arguments:
        config (dict): Parsed section of the YAML configuration file.

    Returns:
        Boolean indicating result and ``Deployer`` instance or ``None``.
    """
    try:
        deployer = Deployer(**config)

    except KeyError as e:
        # Missing required parameter
        key = e.args[0]
        cprint(m.DEP_MISSING_PARAM + '\n' % key, 'red')
        return False, None

    # Determine deployment function to use
    if deployer.source_type == 'local':
        cprint(m.DEP_LOCAL)
        deployer.deploy = types.MethodType(deployments.deploy_local, deployer)

    elif deployer.source_type == 'git':
        cprint(m.DEP_GIT)
        deployer.deploy = types.MethodType(deployments.deploy_git, deployer)

    else:
        # Unknown deployment type
        cprint(m.DEP_UNKNOWN % deployer.source_type, 'red')
        return False, None

    # Additional method for preparing/testing the deployment
    deployer.prepare = types.MethodType(deployments.prepare, deployer)

    return True, deployer



# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implementation of the ``prepare`` "deployment".

This is simply used to check connection configuration and create the remote
directory tree structure.
"""

from fumi import messages as m
from fumi import util


def prepare(deployer):
    """Test connection and prepare directories.

    Arguments:
        deployer (``Deployer``): Deployer instance.

    Returns:
        Boolean indicating result of the preparation.
    """
    # SSH connection
    util.cprint(
        '> ' + m.DEP_CONNECTING % (deployer.host, deployer.user),
        'cyan'
    )

    status, ssh = util.connect(deployer)
    if not status:
        ssh.close()
        return False

    util.cprint(m.DEP_CONNECTED + '\n', 'green')


    # Directory structures
    util.cprint('> ' + m.DEP_CREATEDIR, 'cyan')

    status = util.create_dirs(ssh, deployer)
    if not status:
        ssh.close()
        return False

    util.cprint(m.CORRECT + '\n', 'green')


    util.cprint(m.DEP_PREPARE_COMPLETE, 'green')
    util.cprint(m.DEP_PREPARE_NOTICE, 'white')

    # Close SSH connection
    ssh.close()

    return True



# -*- coding: utf-8 -*-
#
# fumi deployment tool
# https://github.com/rmed/fumi
#
# The MIT License (MIT)
#
# Copyright (c) 2016 Rafael Medina García <rafamedgar@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implementation of the ``local`` based deployment.

This compresses the files to deploy and uploads them to the remote host.
"""

import datetime
import os
import scp
import tarfile

from fumi import messages as m
from fumi import util


def deploy(deployer):
    """Local based deployment.

    Arguments:
        deployer (``Deployer``): Deployer instance.

    Returns:
        Boolean indicating result of the deployment.
    """
    # SSH connection
    util.cprint(
        '> ' + m.DEP_CONNECTING % (deployer.host, deployer.user),
        'cyan'
    )

    status, ssh = util.connect(deployer)
    if not status:
        ssh.close()
        return False

    util.cprint(m.DEP_CONNECTED + '\n', 'green')


    # Predeployment commands
    status, util.run_commands(ssh, deployer.predep)
    if not status:
        ssh.close()
        return False


    # Directory structures
    util.cprint('> ' + m.DEP_CHECK_REMOTE, 'cyan')

    status = util.check_dirs(ssh, deployer)
    if not status:
        ssh.close()
        return False

    util.cprint(m.CORRECT + '\n', 'green')


    # Compress source to temporary directory
    timestamp = datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S')
    util.cprint(m.DEP_PREPARE_REV % timestamp, 'white')

    compressed_file = timestamp + '.tar.gz'
    tmp_local = os.path.join('/tmp', compressed_file)

    if deployer.local_ignore:
        cnt_list = list(
            set(os.listdir(deployer.source_path))^set(deployer.local_ignore))

    else:
        cnt_list = os.listdir(deployer.source_path)

    util.cprint('> ' + m.DEP_LOCAL_COMPRESS % tmp_local, 'cyan')

    with tarfile.open(tmp_local, "w:gz") as tar:
        for item in cnt_list:
            path = os.path.join(deployer.source_path, item)

            if os.path.exists(path):
                # Compress
                tar.add(path, arcname=timestamp + '/' + item)

            else:
                # Ignore
                util.cprint(m.DEP_LOCAL_PATHNOEXIST % path, 'white')

    util.cprint(m.DONE + '\n', 'green')


    # Upload compressed source
    util.cprint('> ' + m.DEP_LOCAL_UPLOAD % compressed_file, 'cyan')

    try:
        uload = scp.SCPClient(
            ssh.get_transport(),
            buff_size=deployer.buffer_size)

    except:
        # Failed to initiate SCP
        util.cprint(m.DEP_LOCAL_SCPFAIL, 'red')
        status = util.rollback(ssh, deployer, timestamp, 1)
        ssh.close()
        return False

    uload_tmp = deployer.host_tmp or '/tmp'
    uload_path = os.path.join(uload_tmp, compressed_file)

    try:
        uload.put(tmp_local, uload_path)

    except scp.SCPException as e:
        util.cprint(m.DEP_LOCAL_UPLOADERR + '\n' % e, 'red')
        status = util.rollback(ssh, deployer, timestamp, 3)
        ssh.close()
        return False

    util.cprint(m.DONE + '\n', 'green')


    # Uncompress source to deploy_path/rev
    util.cprint('> ' + m.DEP_LOCAL_UNCOMPRESS, 'cyan')

    rev_path = os.path.join(deployer.deploy_path, 'rev')
    untar = 'tar -C %s -zxvf %s' % (rev_path, uload_path)

    stdin, stdout, stderr = ssh.exec_command(untar)
    status = stdout.channel.recv_exit_status()

    if status == 127:
        util.cprint(m.DEP_LOCAL_ERR127 + '\n', 'red')

        status = util.rollback(ssh, deployer, timestamp, 1)
        ssh.close()
        return False

    elif status == 1:
        util.cprint(m.DEP_LOCAL_ERR1 + '\n', 'red')
        util.cprint(*stderr.readlines())

        status = util.rollback(ssh, deployer, timestamp, 2)
        ssh.close()
        return False

    elif status == 2:
        util.cprint(m.DEP_LOCAL_ERR2 + '\n', 'red')
        util.cprint(*stderr.readlines())

        status = util.rollback(ssh, deployer, timestamp, 2)
        ssh.close()
        return False

    util.cprint(m.DONE + '\n', 'green')


    # Link directory
    status = util.symlink(ssh, deployer, rev_path, timestamp)
    if not status:
        util.rollback(ssh, deployer, timestamp, 3)
        ssh.close()
        return False


    # Link shared paths
    util.symlink_shared(ssh, deployer)


    # Run post-deployment commands
    status = util.run_commands(
        ssh,
        deployer.postdep,
        os.path.join(deployer.deploy_path, 'current'))

    if not status:
        util.rollback(ssh, deployer, timestamp, 3)
        ssh.close()
        return False


    # Clean revisions
    if deployer.keep_max:
        status = util.clean_revisions(ssh, deployer.keep_max, rev_path)


    # Cleanup temporary files
    util.cprint('> ' + m.DEP_LOCAL_CLEAN, 'cyan')

    util.remove_local(tmp_local)
    util.remove_remote(ssh, uload_path)

    util.cprint(m.DONE + '\n', 'green')


    util.cprint(m.DEP_COMPLETE, 'green')

    # Close SSH connection
    ssh.close()

    return True
